////////////////////////////////////////////////////////////////////////
// Class:       RawWaveformAna
// Plugin Type: analyzer (art v3_05_01)
// File:        RawWaveformAna_module.cc
//
// Generated at Mon May 25 10:53:11 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
////////////////////////////////////////////////////////////////////////

// art framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"

// larsoft includes
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardata/Utilities/LArFFT.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcore/Geometry/Geometry.h"

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

// ROOT includes
#include "TH1.h"
#include "TH2.h"
#include "TTree.h"
#include "TFile.h"
#include "TCanvas.h"
#include "TProfile.h"


using namespace std;

class RawWaveformAna;


class RawWaveformAna : public art::EDAnalyzer {
public:
  explicit RawWaveformAna(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  RawWaveformAna(RawWaveformAna const&) = delete;
  RawWaveformAna(RawWaveformAna&&) = delete;
  RawWaveformAna& operator=(RawWaveformAna const&) = delete;
  RawWaveformAna& operator=(RawWaveformAna&&) = delete;

  // Required functions.
  void analyze(art::Event const& e) override;
  
  void beginJob() override;

private:

  // Declare member data here.
  
  // user's function
  void reset();
  void calculateFFT(TH1D* h1_waveform, TH1D* h1_fft);
  std::vector<float> getMedianCorrection(std::vector< std::vector<float> >& holder, unsigned int dataSize, unsigned short plane);
  std::vector<bool> roiMask(std::vector<float>& channel, unsigned short plane);


  // configurtion from fcl file
  std::string fRawDigitLabel;
  std::string fRawDigitInstance;
  std::vector<double> fRoiThresholdLow; // threshold to check sig/noise low
  std::vector<double> fRoiThresholdHigh; // threshold to check sig/noise high
  int fRoiPadLow; // signal-like region
  int fRoiPadHigh; // signal-like region
  bool fApplyCoherentNoiseRemoving;
  
  // TTree
     TTree *fEventTree;

  // event information
  int event;
  int run;
  int subrun;

  // TPC information
  int fNticks;
  int fNticksReadout;
  float fSampleRate;

  float fFFTBinWidth;
  std::vector<TH2F*> fDisplay; // [plane]
  std::vector<TH2F*> fDisplay_cohNoiseRemoved; // [plane]
  std::vector<TH2I*> fChannelRawFFT; // [plane]
  std::vector<TProfile*> fMagRawFFT_sig_pfx; // [plane]
  std::vector<TProfile*> fMagRawFFT_noise_pfx; // [plane]
  std::vector<TProfile*> fMagRawFFT_sig_pfx_cohNoiseRemoved; // [plane]
  std::vector<TProfile*> fMagRawFFT_noise_pfx_cohNoiseRemoved; // [plane]


  art::ServiceHandle<geo::Geometry const> fGeometry;
};


RawWaveformAna::RawWaveformAna(fhicl::ParameterSet const& p)
  : EDAnalyzer{p}  // ,
  // More initializers here.
{
  // Call appropriate consumes<>() for any products to be retrieved by this module.
  fRawDigitLabel    = p.get<std::string>("RawDigitLabel");
  fRawDigitInstance    = p.get<std::string>("RawDigitInstance");
  fRoiThresholdLow    = p.get<std::vector<double>>("RoiThresholdLow");
  fRoiThresholdHigh    = p.get<std::vector<double>>("RoiThresholdHigh");
  fRoiPadLow    = p.get<int>("RoiPadLow");
  fRoiPadHigh    = p.get<int>("RoiPadHigh");
  fApplyCoherentNoiseRemoving   = p.get<bool>("ApplyCoherentNoiseRemoving");


  // DetectorPropertiesService
  auto const *detprop = lar::providerFrom<detinfo::DetectorPropertiesService>();
  fNticks = detprop->NumberTimeSamples(); // number of clock ticks per event
  fNticksReadout = detprop->ReadOutWindowSize(); // number of clock ticks per readout window
  fSampleRate = detprop->SamplingRate(); // period of the TPC readout electronics clock [us]; 1 tick = 1 tdc = 198 ns.
  cout << "Numer of clock ticks per event: " << fNticks << endl;
  cout << "Numer of clock ticks per readout window: " << fNticksReadout << endl;
  cout << "Sampling rate: " << fSampleRate << endl;
  // width of frequency bin in kHz (signal window is 2048 here)
  fFFTBinWidth = 1.0/(fNticks*fSampleRate*1.0e-6); // [kHz]
  cout << "fFFTBinWidth: " << fFFTBinWidth << endl;
}

void RawWaveformAna::analyze(art::Event const& e)
{
  // Implementation of required member function here.
  reset();

  // event info
  run = e.run();
  subrun = e.subRun();
  event = e.id().event();

  // channel status
  lariov::ChannelStatusProvider const& channelStatus
    = art::ServiceHandle<lariov::ChannelStatusService const>()->GetProvider();

  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, fRawDigitInstance, rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }

  // ------ coherent noise ---------------
  // divide channels into 20 groups, find noise ADC median among channels in group for each tick (there are 2048 ticks in total) 
  // the noise ADC median is called coherent noise

  unsigned int nchgrps = 20; // channel groups
  unsigned int nchpergrp = 24; // channels per group

  std::vector< std::vector<float> > cohNoise; // 0-10 induction, 11-20, collection; [20][2048]
  
  // loop through groups of channels
  for (unsigned int chgrp = 0; chgrp < nchgrps; ++chgrp) {
    std::vector< std::vector<float> > channelgroup; // [channels][2048]: exclude bad channels in the group, channels <=24

    unsigned short plane = 0;
    unsigned int dataSize = fNticks;
    
    // loop through channels, 24 channels/group
    for (unsigned int ch = 0; ch < nchpergrp; ++ch) {
      unsigned int channel = chgrp*nchpergrp + ch;

      if(channelStatus.IsBad(channel)) continue;

      if (channel >= 240) plane = 1;

      art::Ptr<raw::RawDigit> digit(rawdigitListHandle, channel);
      dataSize = digit->Samples(); // 2048
      raw::RawDigit::ADCvector_t rawadc(dataSize);  // vector holding uncompressed adc values
      raw::Uncompress(digit->ADCs(), rawadc, digit->Compression());

      float ped = digit->GetPedestal();
      std::vector<float> holder(dataSize, 0.);
      for(unsigned short bin = 0; bin < dataSize; ++bin) {
        holder[bin] = rawadc[bin] - ped;
      }

      channelgroup.push_back(holder);
    } // loop through nchpergrp ch

    auto correction = getMedianCorrection(channelgroup, dataSize, plane);
    cohNoise.push_back(correction);


  } // loop over nchgrps chgrp


  // ------- extract noise & fft----------------
  int countGoodChannels = 0;
  // loop over all RawDigit (entire channels)
  for (size_t i=0; i<rawdigitlist.size(); ++i) {
    art::Ptr<raw::RawDigit> rawdigit = rawdigitlist[i];

    // channel number
    int channel = rawdigit->Channel();
    if (channelStatus.IsBad(channel)) continue;
    countGoodChannels++;

    auto plane = fGeometry->ChannelToWire(channel)[0].Plane; // ChannelToWire: Returns a list of wires connected to the specified TPC channel. 
    //auto view = fGeometry->View(channel); // same as plane

    // number of samples in uncompressed ADC data
    int nSamples =  rawdigit->Samples();

    // pedestal
    int pedestal = rawdigit->GetPedestal();
    //cout << "rawdigit->GetPedestal(): " << rawdigit->GetPedestal() << endl;  

    std::vector<short> uncompressed(nSamples);
    // with pedestal
    raw::Uncompress(rawdigit->ADCs(), uncompressed, pedestal, rawdigit->Compression());
    // subtract pedestals
    std::vector<short> uncompPed(nSamples);
    float amp = 0; // maximum amplitude
    for (int j=0; j<nSamples; j++) {
      uncompPed.at(j)=uncompressed.at(j)-pedestal;
      if (uncompPed.at(j) > amp) amp = uncompPed.at(j);
    }

    // subtract coherent noise
    std::vector<float> uncompPed_cohNoiseRemoved(nSamples);
    unsigned int chgroup = floor(channel / nchpergrp);
    for (int j=0; j<nSamples; j++) {
      uncompPed_cohNoiseRemoved[j] = uncompPed[j] - cohNoise[chgroup][j];
    }

    // waveform histogram
    TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), nSamples, 0, nSamples); 
    TH1D* h1_waveform_cohNoiseRemoved = new TH1D(Form("waveform_cohNoiseRemoved_%d", channel), Form("waveform_cohNoiseRemoved_%d", channel), nSamples, 0, nSamples); 
    for (int k=0; k<nSamples; k++) {
      h1_waveform->SetBinContent(k+1, uncompPed.at(k));
      h1_waveform_cohNoiseRemoved->SetBinContent(k+1, uncompPed_cohNoiseRemoved.at(k));
    }

    // do FFT for single waveform
    TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), nSamples, 0, nSamples*fFFTBinWidth);
    TH1D* h1_fft_cohNoiseRemoved = new TH1D(Form("fft_cohNoiseRemoved_%d", channel), Form("fft_cohNoiseRemoved_%d", channel), nSamples, 0, nSamples*fFFTBinWidth);

    calculateFFT(h1_waveform, h1_fft);
    calculateFFT(h1_waveform_cohNoiseRemoved, h1_fft_cohNoiseRemoved);

    // save for each plane
    //for (int k=0; k<nSamples/2; k++) {
    for (int k=0; k<nSamples; k++) {
      fDisplay[plane]->Fill(channel, k+0.5, h1_waveform->GetBinContent(k+1));
      fDisplay_cohNoiseRemoved[plane]->Fill(channel, k+0.5, h1_waveform_cohNoiseRemoved->GetBinContent(k+1));
      if (h1_fft->GetBinContent(k+1) != 0) {
        fChannelRawFFT[plane]->Fill(channel, (k+0.5)*fFFTBinWidth, h1_fft->GetBinContent(k+1));
      }
    }

    //for (int k=0; k<nSamples/2; k++) {
    for (int k=0; k<nSamples; k++) {
      if (amp > fRoiThresholdHigh[plane]) {
        fMagRawFFT_sig_pfx[plane]->Fill((k+0.5)*fFFTBinWidth, h1_fft->GetBinContent(k+1), 1);
        fMagRawFFT_sig_pfx_cohNoiseRemoved[plane]->Fill((k+0.5)*fFFTBinWidth, h1_fft_cohNoiseRemoved->GetBinContent(k+1), 1);
      }
      else {
        fMagRawFFT_noise_pfx[plane]->Fill((k+0.5)*fFFTBinWidth, h1_fft->GetBinContent(k+1), 1);
        fMagRawFFT_noise_pfx_cohNoiseRemoved[plane]->Fill((k+0.5)*fFFTBinWidth, h1_fft_cohNoiseRemoved->GetBinContent(k+1), 1);
      }
    }
    
  } // loop over rawdigitlist.size() i

  //cout << "countGoodChannels: " << countGoodChannels << endl;

  // fill ttree
  fEventTree->Fill();
}

void RawWaveformAna::beginJob(){
  // ttree
  art::ServiceHandle<art::TFileService> tfs;
  fEventTree = tfs->make<TTree>("Event", "Event");
  fEventTree->Branch("event", &event, "event/I");
  fEventTree->Branch("run", &run, "run/I");
  fEventTree->Branch("subrun", &subrun, "subrun/I");

  // argoneut has one cryostat, one tpc, two planes and 480 channels with 240 channels for each plane
  for (unsigned int p=0; p<fGeometry->Nplanes(); p++) {
    TH2F* fPlaneDisplay = tfs->make<TH2F>(Form("h2_display_plane_%d",p), Form("Raw event display (plane %d); Channel; Tick",p), fGeometry->Nchannels()/2, 0+p*fGeometry->Nchannels()/2, fGeometry->Nchannels()/2-1+p*fGeometry->Nchannels()/2, fNticks, 0, fNticks);
    fDisplay.push_back(fPlaneDisplay);

    TH2F* fPlaneDisplay_cohNoiseRemoved = tfs->make<TH2F>(Form("h2_display_cohNoiseRemoved_plane_%d",p), Form("Raw event display after removing coherent noise (plane %d); Channel; Tick",p), fGeometry->Nchannels()/2, 0+p*fGeometry->Nchannels()/2, fGeometry->Nchannels()/2-1+p*fGeometry->Nchannels()/2, fNticks, 0, fNticks);
    fDisplay_cohNoiseRemoved.push_back(fPlaneDisplay_cohNoiseRemoved);

    TH2I* fPlaneChannelRawFFT = tfs->make<TH2I>(Form("ChannelRawFFT_plane_%d",p), Form("RawFFT ADC vs Channel (plane %d); Channel; ADC RawFFT",p), fGeometry->Nchannels()/2, 0+p*fGeometry->Nchannels()/2, fGeometry->Nchannels()/2-1+p*fGeometry->Nchannels()/2, fNticks, 0, fNticks);
    fChannelRawFFT.push_back(fPlaneChannelRawFFT);

    TProfile* fPlaneMagRawFFT_sig_pfx = tfs->make<TProfile>(Form("fMagRawFFT_sig_pfx_plane_%d",p), Form("Profiled RawFFT (plane %d); Frequency [kHz]; Magnitude",p), fNticks, 0, fNticks*fFFTBinWidth);
    fMagRawFFT_sig_pfx.push_back(fPlaneMagRawFFT_sig_pfx);

    TProfile* fPlaneMagRawFFT_sig_pfx_cohNoiseRemoved = tfs->make<TProfile>(Form("fMagRawFFT_sig_pfx_cohNoiseRemoved_plane_%d",p), Form("Profiled RawFFT (plane %d); Frequency [kHz]; Magnitude",p), fNticks, 0, fNticks*fFFTBinWidth);
    fMagRawFFT_sig_pfx_cohNoiseRemoved.push_back(fPlaneMagRawFFT_sig_pfx_cohNoiseRemoved);

    TProfile* fPlaneMagRawFFT_noise_pfx = tfs->make<TProfile>(Form("fMagRawFFT_noise_pfx_plane_%d",p), Form("Profiled RawFFT (plane %d); Frequency [kHz]; Magnitude",p), fNticks, 0, fNticks*fFFTBinWidth);
    fMagRawFFT_noise_pfx.push_back(fPlaneMagRawFFT_noise_pfx);

    TProfile* fPlaneMagRawFFT_noise_pfx_cohNoiseRemoved = tfs->make<TProfile>(Form("fMagRawFFT_noise_pfx_cohNoiseRemoved_plane_%d",p), Form("Profiled RawFFT (plane %d); Frequency [kHz]; Magnitude",p), fNticks, 0, fNticks*fFFTBinWidth);
    fMagRawFFT_noise_pfx_cohNoiseRemoved.push_back(fPlaneMagRawFFT_noise_pfx_cohNoiseRemoved);
  }
  
}

void RawWaveformAna::calculateFFT(TH1D* h1_waveform, TH1D* h1_fft) {
  int nbins = h1_waveform->GetNbinsX();
  
  TH1* h1_transform = 0;
  
  h1_transform = h1_waveform->FFT(h1_transform, "MAG");

  // loop over h1_transform
  for (int k=0; k<nbins; k++) {
    double frequency = (k+0.5) * fFFTBinWidth; // [kHz]
    double amplitude = h1_transform->GetBinContent(k+1);
    h1_fft->Fill(frequency, amplitude);
  }

  h1_transform->Delete();
}

std::vector<float> RawWaveformAna::getMedianCorrection(std::vector< std::vector<float> >& channelgroup, unsigned int dataSize, unsigned short plane)
{ 
  std::vector<float> correction(dataSize);
  std::vector< std::vector<float> > samples(dataSize); // [2048][channels in group with noise]

  for(size_t ch = 0; ch < channelgroup.size(); ++ch) {
    std::vector<float> thischannel = channelgroup[ch];
    auto mask = roiMask(thischannel, plane);

    for(size_t s = 0; s < samples.size(); ++s) {
      if (!mask[s]) { continue; }
      samples[s].push_back(thischannel[s]);
    }
  }

  for(size_t s = 0; s < samples.size(); ++s) {
    size_t n = samples[s].size();
    if(n < 2) { correction[s] = 0; continue; }
    std::sort(samples[s].begin(), samples[s].end());
    if((n % 2) == 0) { correction[s] = 0.5 * (samples[s][n/2] + samples[s][(n/2)-1]); }
    else { correction[s] = samples[s][(n-1)/2]; }
  }

  return correction;
} // getMedianCorrection

std::vector<bool> RawWaveformAna::roiMask(std::vector<float>& channel, unsigned short plane)
{
  // mask: true -> noise-like; false -> signal-like
  std::vector<bool> mask(channel.size(), true); // [2048]

  bool inroi = false;
  for (int i = 0; i < (int)channel.size(); ++i) {
    
    auto sig = channel[i]; // ADC at tick i

    if (inroi){
      if (sig > fRoiThresholdHigh[plane] || sig < -1.0*fRoiThresholdHigh[plane]) { 
        mask[i] = false; // signal-like
      }
      else {
        for (int p = 0; p <= fRoiPadHigh; ++p) {
          if ((i + p) < (int)mask.size())  mask[i + p] = false;
        }
        inroi = false;
      }   
    } // if (inroi)
    else {
      if (sig > fRoiThresholdLow[plane] || sig < -1.0*fRoiThresholdLow[plane]) {
        for (int p = fRoiPadLow; p >= 0; --p) {
          if (i - p >= 0) mask[i - p] = false;
        }
        inroi = true;
      }   
    }
  } // loop over channel.size() i

  return mask;
} // roiMask


void RawWaveformAna::reset(){
  run = -9999;
  subrun = -9999;
  event = -9999;
}

DEFINE_ART_MODULE(RawWaveformAna)
